"""
Testing cmd
"""

# pylint: disable=line-too-long,missing-function-docstring,missing-class-docstring,super-with-arguments,invalid-name,R0801

import os
import subprocess
import unittest
from pprint import pprint

HERE = os.path.dirname(os.path.abspath(__file__))
FETCH_LIST = os.path.join(HERE, "fetch_list.json")
OUT_LIST = os.path.join(HERE, "out_list.json")


class IoTester(unittest.TestCase):
    def test_env(self):  # pylint: disable
        pprint(dict(os.environ))

    def test_cmd(self):  # pylint: disable
        subprocess.check_call(
            "vidcrawler --help",
            shell=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

    def test_process_test(self):  # pylint: disable
        subprocess.check_call(
            f'vidcrawler --input_crawl_json "{FETCH_LIST}" --output_json "{OUT_LIST}"',
            shell=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import os
import unittest

from vidcrawler.spotify import fetch_spotify_today

RUN_ALL_TESTS = os.environ.get("RUN_ALL_TESTS", None) is not None


class SpotifyScraperTest(unittest.TestCase):
    @unittest.skipUnless(RUN_ALL_TESTS, "Spotify is not working")
    def test_joe_rogan(self):
        vid_list = fetch_spotify_today(channel_name="Joe Rogan", channel="4rOoJ6Egrf8K2IrywzwOMk")
        self.assertTrue(vid_list)


if __name__ == "__main__":
    unittest.main()



"""Spider which crawls all the channels"""

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return

import json
import queue
import random
import signal
import sys
import threading
import time
import traceback
from typing import Any, Dict, List, Tuple

from .bitchute import fetch_bitchute_today
from .brighteon import fetch_brighteon_today
from .gabtv import fetch_gabtv_today
from .odysee import fetch_odysee_today
from .rumble import fetch_rumble_channel_today
from .spotify import fetch_spotify_today
from .spreaker import fetch_spreaker_today
from .video_info import VideoInfo
from .youtube import fetch_youtube_today

BRIGHTEON = "brighteon"
YOUTUBE = "youtube"
RUMBLE = "rumble"
BITCHUTE = "bitchute"
SPREAKER = "spreaker"
ODYSEE = "odysee"
GAB = "gab"
SPOTIFY = "spotify"

CRAWLER_MAP = {
    RUMBLE: fetch_rumble_channel_today,
    BITCHUTE: fetch_bitchute_today,
    BRIGHTEON: fetch_brighteon_today,
    YOUTUBE: fetch_youtube_today,
    SPREAKER: fetch_spreaker_today,
    ODYSEE: fetch_odysee_today,
    GAB: fetch_gabtv_today,
    SPOTIFY: fetch_spotify_today,
}

_SCRAPE_RANDOMIZE_ORDER = True
_YT_SCRAPER_ENABLE_THREADS = True
_YT_SCRAPER_ENABLE_THREADS = False

# type: ignore


def _fetch_all(
    source_name: str,
    fetch_list: List[Tuple[str, str]],
    out_queue: queue.Queue,
    bad_channels: queue.Queue,
) -> None:
    callback = CRAWLER_MAP[source_name]
    if _SCRAPE_RANDOMIZE_ORDER:
        random.shuffle(fetch_list)

    def do_fetch(fetch_list: List, callback: Any) -> None:
        for channel_name, channel_id in fetch_list:
            try:
                videos = callback(channel_name, channel_id)  # type: ignore
                for vid in videos:
                    out_queue.put(vid)
            except KeyboardInterrupt:
                sys.exit(1)
            except BaseException as e:
                traceback.print_exc()
                entry = (channel_name, str(e))
                bad_channels.put(entry)
                continue

    if source_name != YOUTUBE:
        do_fetch(fetch_list, callback)
        return
    if not _YT_SCRAPER_ENABLE_THREADS:
        do_fetch(fetch_list, callback)
        return
    # YouTube and threads enabled.
    num_workers = 4
    thread_data = [[] for _ in range(num_workers)]  # type: ignore
    for i, data in enumerate(fetch_list):
        idx = i % num_workers
        thread_data[idx].append(data)
    threads = []
    for tdata in thread_data:
        thread = threading.Thread(target=do_fetch, args=(tdata, callback))
        thread.daemon = True
        thread.start()
        threads.append(thread)
    for t in threads:
        t.join()


def _yield_thread_for_keyboard_interrupt() -> None:
    if sys.platform == "win32":
        time.sleep(0.1)
    else:
        try:
            # might be missing on some platforms.
            signal.pause()  # pylint: disable-all
        except BaseException:
            time.sleep(0.1)


def _threaded_fetch_channels(
    channels: List[Tuple[str, str, str]],
    out_videos: List[VideoInfo],
    out_bad_channels: List[Tuple[str, str]],
) -> None:
    # Partition the list by source type (bitchute, youtube, ...)
    # for each source fetch all videos.
    partitioned: Dict[str, Any] = {}
    for channel_name, source, channel_id in channels:
        partitioned.setdefault(source, [])
        partitioned[source].append((channel_name, channel_id))
    queue_out: queue.Queue = queue.Queue()
    queue_bad_channels: queue.Queue = queue.Queue()
    # Launch threads but also makes the main thread responsive to the keyboard interrupt.
    threads = []
    for source, array_data in partitioned.items():
        thread = threading.Thread(
            target=_fetch_all,
            args=(source, array_data, queue_out, queue_bad_channels),
        )
        thread.daemon = True
        thread.start()
        threads.append(thread)
    while True:  # Weird loop allows KeyboardInterrupt
        finish_cnt = 0
        for t in threads:
            if t.is_alive():
                finish_cnt += 1
        if finish_cnt == len(threads):
            break
        _yield_thread_for_keyboard_interrupt()
    # Busy loop here allows keyboard interrupts to filter through to the
    # top and halt the execution of the program right away.
    while len(threads):
        for i, thread in enumerate(threads):
            if not thread.is_alive():
                thread.join()
                threads.pop(i)
                break  # Start the outer loop again
        time.sleep(0.1)
    while not queue_out.empty():
        out_videos.append(queue_out.get_nowait())
    for c in list(queue_bad_channels.queue):
        out_bad_channels.append(c)


def _singlethreaded_fetch(
    channels: List[Tuple[str, str, str]],
    out_videos: List[VideoInfo],
    out_bad_channels: List[Tuple[str, str]],
) -> None:
    partitioned: Dict[str, Any] = {}
    for channel_name, source, channel_id in channels:
        partitioned.setdefault(source, [])
        partitioned[source].append((channel_name, channel_id))
    queue_out: queue.Queue = queue.Queue()
    queue_bad_channels: queue.Queue = queue.Queue()
    for source, array_data in partitioned.items():
        _fetch_all(source, array_data, queue_out, queue_bad_channels)
    while not queue_out.empty():
        out_videos.append(queue_out.get_nowait())
    for c in list(queue_bad_channels.queue):
        out_bad_channels.append(c)


def crawl_video_sites(channels: List[Tuple[str, str, str]], use_threads: bool = True) -> str:  # type: ignore
    vid_infos: List[VideoInfo] = []
    bad_channels: List[Tuple[str, str]] = []
    if use_threads:
        _threaded_fetch_channels(channels, vid_infos, bad_channels)
    else:
        _singlethreaded_fetch(channels, vid_infos, bad_channels)
    # apply_fetch_images(vid_infos)
    out_data: List[Dict] = VideoInfo.to_plain_list(vid_infos)  # type: ignore
    json_str = json.dumps(out_data, indent=2, sort_keys=True, ensure_ascii=False)
    bad_channels.sort()
    if bad_channels:
        print("#############")
        print("# Bad channels:")
        for name, err in bad_channels:
            print(f"#  {name}: {err}")
        print("#############")
    else:
        print("No bad channels detected!")
    return json_str



"""
Launches playwright and sets up all the defaults.
"""

import os
from contextlib import contextmanager
from typing import Generator

from filelock import FileLock
from playwright.sync_api import Browser, Page, sync_playwright

INSTALLED = False
HEADLESS = True


def set_headless(headless: bool) -> None:
    """Set headless."""
    global HEADLESS  # pylint: disable=global-statement
    HEADLESS = headless


def install_playwright() -> None:
    """Install Playwright."""
    install_lock = os.path.join(os.getcwd(), "playwright.lock")
    with FileLock(install_lock):
        global INSTALLED  # pylint: disable=global-statement
        if INSTALLED:
            return
        rtn = os.system("playwright install")
        if rtn != 0:
            raise OSError("Failed to install Playwright.")
        INSTALLED = True



@contextmanager
def launch_playwright() -> Generator[tuple[Page, Browser], None, None]:
    """
    Launches a playwright browser. Each browser is only safe to use in a single thread.
    """
    install_playwright()
    with sync_playwright() as context:
        headless = HEADLESS or os.environ.get("GITHUB_ACTIONS") == "true"
        browser = context.chromium.launch(headless=headless)
        page = browser.new_page()
        try:
            yield (page, browser)
        finally:
            page.close()
            browser.close()



"""
Simple Http Server useful for testing
"""

# pylint: disable=line-too-long,missing-function-docstring,missing-class-docstring,super-with-arguments

import atexit
import http.server
import socketserver
import threading
from contextlib import contextmanager
from http import HTTPStatus
from typing import Any, Callable, Generator, List, Optional

import requests

StringFunctor = Callable[[], str]

_TIMEOUT = 10


def make_handler_class(response_text_fcn: StringFunctor) -> Any:
    class Handler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self):
            self.send_response(HTTPStatus.OK)
            self.end_headers()
            data = response_text_fcn().encode("utf-8")
            self.wfile.write(data)

    return Handler


class ServerThread(threading.Thread):
    def __init__(self, tcp_server: socketserver.TCPServer):
        threading.Thread.__init__(self)
        self.tcp_server: socketserver.TCPServer = tcp_server
        _alive_servers.append(self)

    def run(self):
        self.tcp_server.serve_forever(poll_interval=0.001)

    def join(self, timeout: Optional[float] = 2) -> None:
        _alive_servers.remove(self)
        self.tcp_server.shutdown()
        self.tcp_server.server_close()
        super(ServerThread, self).join(timeout=timeout)


_alive_servers: List[ServerThread] = []


@atexit.register
def _exit_dead_servers():
    """Just incase the unit tests misbehave, this will clean up threads."""
    for server in _alive_servers:
        server.join()


@contextmanager
def simple_response_server_thread(
    port: int,
    response_text_fcn: StringFunctor = lambda: "Hello World!",
) -> Generator[ServerThread, None, None]:
    """
    Example:
        with simple_response_server_thread(port=53925, response_text_fcn=lambda: "this should match!!!!"):
            resp = requests.get("http://localhost:53925")
            resp.raise_for_status()
            assert resp.text == "this should match!!!!"
    """
    handler_class = make_handler_class(response_text_fcn=response_text_fcn)
    tcp_server: socketserver.TCPServer = socketserver.TCPServer(("", port), handler_class)
    tcp_server.allow_reuse_address = True

    try:
        server_thread = ServerThread(tcp_server)
        server_thread.start()
        yield server_thread
    finally:
        server_thread.join()


def unit_test() -> None:
    with simple_response_server_thread(port=53925, response_text_fcn=lambda: "this should match!!!!"):
        resp = requests.get("http://localhost:53925", timeout=_TIMEOUT)
        resp.raise_for_status()
        assert resp.text == "this should match!!!!"


if __name__ == "__main__":
    unit_test()



"""
Rumble scrapper.
"""

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return,fixme

import json
import sys
import traceback
from typing import Dict, List, Tuple

import isodate  # type: ignore
import requests
from bs4 import BeautifulSoup  # type: ignore

from .date import iso_fmt, now_local
from .fetch_html import fetch_html_using_request_lib as fetch_html
from .video_info import VideoInfo


def parse_rumble_video_object(html_doc: str) -> Dict[str, str]:
    soup_article = BeautifulSoup(html_doc, "html.parser")
    script_dom = soup_article.find("script", {"type": "application/ld+json"})
    json_string = script_dom.contents[0]
    data = json.loads(json_string)  # type: ignore
    return data[0]  # type: ignore


def fetch_rumble(channel: str) -> Tuple[str, str]:
    html_doc: str = ""
    channel_url: str = "https://rumble.com/c/%s" % channel
    url: str = "https://rumble.com/c/%s?date=this-month" % channel
    try:
        sys.stdout.write("Rumble visiting %s (%s)\n" % (channel, url))
        html_doc = fetch_html(url)
    except requests.exceptions.HTTPError:
        channel_url = "https://rumble.com/user/%s" % channel
        url = "https://rumble.com/user/%s" % channel
        sys.stdout.write(f"Rumble alt visiting {channel} ({url})\n")
        html_doc = fetch_html(url)
    return (html_doc, channel_url)


# type: ignore


def fetch_rumble_channel_today(channel_name: str, channel: str) -> List[VideoInfo]:
    # TODO: Fine grained try...catch blocks.
    output: List[VideoInfo] = []
    html_doc: str = ""
    channel_url: str = ""
    html_doc, channel_url = fetch_rumble(channel)
    # print(html_doc)
    soup = BeautifulSoup(html_doc, "html.parser")
    for article in soup.find_all("article", class_="video-item"):
        try:
            article_duration_dom = article.find(class_="video-item--duration")
            duration = "" if not article_duration_dom else article_duration_dom["data-value"]
            vid_src_suffix = article.find(class_="video-item--a")["href"]
            vid_src = "https://rumble.com%s" % vid_src_suffix
            sys.stdout.write("  visiting video %s (%s)\n" % (channel, vid_src))
            html_doc2 = fetch_html(vid_src)
            video_obj = parse_rumble_video_object(html_doc2)
            title = video_obj["name"]
            iframe_src = video_obj["embedUrl"]
            desc_text = video_obj["description"]
            publish_date = video_obj["uploadDate"]

            if not duration:
                duration_str = video_obj["duration"]
                duration = isodate.parse_duration(duration_str).total_seconds()
            try:
                views = video_obj["interactionStatistic"]["userInteractionCount"]  # type: ignore
            except KeyError:
                views = "?"
            # img_src = video_obj['thumbnailUrl']
            img_src = article.find(class_="video-item--img")["src"]
            now_datestr = iso_fmt(now_local())
            o = VideoInfo(
                channel_name=channel_name,
                source="rumble.com",
                date_published=publish_date,
                date_discovered=now_datestr,
                date_lastupdated=now_datestr,
                url=vid_src,
                channel_url=channel_url,
                title=title,
                duration=duration,
                description=desc_text,
                img_src=img_src,
                iframe_src=iframe_src,
                views=views,
                profile_img_src="",
            )
            output.append(o)
        except BaseException as e:  # pylint: disable=broad-except
            s = "".join(traceback.format_exception(None, e, e.__traceback__))
            sys.stdout.write("Error: %s\nCould not parse\n%s\n\n" % (str(s), str(article)))
    return output


def interactive_test() -> None:
    sys.stdout.write("Rumble user: ")
    name: str = input()
    rslt = fetch_rumble_channel_today(channel_name=name, channel=name)
    print(rslt)


if __name__ == "__main__":
    interactive_test()



# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import unittest
from typing import List

from vidcrawler.gabtv import fetch_gabtv_today, fetch_views
from vidcrawler.video_info import VideoInfo


class GabTvTester(unittest.TestCase):
    def test_fetch_gabtv_views(self) -> None:
        """Tests that we can fetch views."""
        views = fetch_views(channel="LadyBee")
        self.assertGreater(len(views), 0)

    def test_fetch_gabtv_today(self) -> None:
        vid_list: List[VideoInfo] = fetch_gabtv_today(channel_name="Maryam Henein", channel_id="LadyBee")
        self.assertGreater(len(vid_list), 0)


if __name__ == "__main__":
    unittest.main()
